"""An Event type for RecurringEvent"""
#
# RecurringEvent/content/event.py
#

__author__  = 'Pawel Marzec <marzec@cjg.pl>'
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo, SpecialUsers
from Products.CMFCore.permissions import ModifyPortalContent, View, ManagePortal
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import ToolNames
from ComputedAttribute import ComputedAttribute

from Products.RecurringEvent.utils import DT2m

from cStringIO import StringIO




try:
  # try to import multilingual support
  from Products.LinguaPlone.public import *
except ImportError:
  # No multilingual support
  from Products.Archetypes.public import *
  

from Products.RecurringEvent.interfaces import IRecurringEvent
from Products.RecurringEvent.config import RECUE_OFFSETS, RECUE_MODES, MONTH_LENGTH 
from Products.RecurringEvent.config import RM_ONE_TIME, RM_UNTIL, RM_FOREVER
from Products.RecurringEvent.config import RO_DAILY, RO_WEEKLY, RO_MONTHLY, RO_YEARLY, FROZEN_DATE_IN_FUTURE
from Products.RecurringEvent.config import ATTENDEE_ROLE, OWNER_ROLE
from Products.RecurringEvent.config import NOTIFY_MSG, PRODID, DATE
from Products.RecurringEvent.config import VCS_HEADER, VCS_FOOTER, VCS_EVENT_START, VCS_EVENT_END
from Products.RecurringEvent.config import ICS_HEADER, ICS_FOOTER, ICS_EVENT_START, ICS_EVENT_END


from Products.ATContentTypes.content.event import ATEvent
from Products.ATContentTypes.content.event import finalizeATCTSchema
from Products.ATContentTypes.permission import ChangeEvents
from Products.ATMemberSelectWidget.ATMemberSelectWidget import *


from DateTime import DateTime

# copy the ATEvent scheme without modifying the original

RecurringEventSchema = ATEvent.schema.copy() + Schema((
	DateTimeField('recueUntilDate',
		required=True,
		searchable=True,
		index="DateTimeIndex:brains",
		write_permission = ChangeEvents,
		default_method=DateTime,
		languageIndependent=True,
		widget = CalendarWidget(
      		description= "",
			description_msgid = "help_event_recue_until",
			label="Event Recue Until",
			label_msgid = "label_event_recue_until",
			i18n_domain = "RecurringEvent")),

	StringField('recueMode',
		required=True,
		searchable=True,
		index="FieldIndex:brains",
		write_permission = ChangeEvents,
		languageIndependent=True,
		vocabulary=RECUE_MODES,
		enforceVocabulary=True,
		default_method="defaultRecueMode",
		widget=SelectionWidget(
			format="select",
			description="choose a recue mode from a list",
			description_msgid='RecurringEvent_help_recueMode',
			i18n_domain='RecurringEvent',
			label="""recue mode""",
			label_msgid='RecurringEvent_label_recueMode')),

	StringField('recueOffset',
		required=True,
		searchable=True,
		index="FieldIndex:brains",
		write_permission = ChangeEvents,
		languageIndependent=True,
		vocabulary=RECUE_OFFSETS,
		enforceVocabulary=True,
		default_method="defaultRecueOffset",
		widget=SelectionWidget(
			format="select",
			description="choose a recue offset from a list",
			description_msgid='RecurringEvent_help_recueOffset',
			i18n_domain='RecurringEvent',
			label="""recue offset""",
			label_msgid='RecurringEvent_label_recueOffset')),

    ComputedField('recueDays',
       index="KeywordIndex:brains",
       read_permisssion = View,
       expression="context.calc_recueDays()"),

    ComputedField('recueWeeks',
       index="KeywordIndex:brains",
       read_permisssion = View,
       expression="context.calc_recueWeeks()"),	

    ComputedField('recueMonths',
       index="KeywordIndex:brains",
       read_permisssion = View,
       expression="context.calc_recueMonths()"),

    ComputedField('recueYears',
       index="KeywordIndex:brains",
       read_permisssion = View,
       expression="context.calc_recueYears()"),
       
    StringField('attendees',
       multiValued=1,
       widget=MemberSelectWidget(fieldType="email")),
       
    BooleanField('onlyForAttendees',),))

# Finalise the schema according to ATContentTypes standards. This basically
# moves the Related items and Allow discussion fields to the bottom of the
# form. See ATContentTypes.content.schemata for details.
finalizeATCTSchema(RecurringEventSchema)

# Declare our class to be specialized event

## class RecurringEvent(ATEvent):
class RecurringEvent(ATEvent):
    """
    An event which can be used to create specialized recurring events with additional
    information about recue.
    """

    # Standard content type setup
    portal_type = meta_type = 'RecurringEvent'
    archetype_name = 'Recurring event'
    content_icon = 'recurring_event.gif'
    schema = RecurringEventSchema
    typeDescription= 'An event which can be used to create specialized recurring events with additional information about recue.'
    typeDescMsgId  = 'RecurringEvent_description_edit'


    # Set up our views - these are available from the 'display' menu
    default_view = 'recurring_event_view'
    immediate_view = 'recurring_event_view'
    # suppl_views = ()

    # Make sure we get title-to-id generation when an object is created
    _at_rename_after_creation = True

    #_atct_newTypeFor = {'portal_type' : 'CMF Event', 'meta_type' : 'CMF Event'}
    #assocMimetypes = ()
    #assocFileExt   = ('revent', )
    #cmf_edit_kws   = ('effectiveDay', 'effectiveMo', 'effectiveYear',
    #                  'expirationDay', 'expirationMo', 'expirationYear',
    #                  'start_time', 'startAMPM', 'stop_time', 'stopAMPM',
    #                  'start_date', 'end_date', 'contact_name', 'contact_email',
    #                  'contact_phone', 'event_url')

    __implements__ = (ATEvent.__implements__, IRecurringEvent,)

    security       = ClassSecurityInfo()

    # actions = updateActions(ATCTContent, CalendarSupportMixin.actions + HistoryAwareMixin.actions)


    #
    # accessors for few default values 
    # ******************************************************************
    
    security.declareProtected(View, 'defaultRecueMode')
    def defaultRecueMode(self):
        return RECUE_MODES.items()[0]


    security.declareProtected(View, 'defaultRecueOffset')
    def defaultRecueOffset(self):
        return RECUE_OFFSETS.items()[0]



    #
    # shifting arithmetic for calculation of recurring
    # ******************************************************************

    security.declareProtected(View, 'numberOfDaysInMonth')
    def numberOfDaysInMonth(self, dt):
       if dt.isLeapYear(): return MONTH_LENGTH['leapYear'][dt.month()]
       return MONTH_LENGTH['standardYear'][dt.month()]


    security.declareProtected(View, 'numberOfDaysInYear')
    def numberOfDaysInYear(self, dt):
        if dt.isLeapYear(): return 366
        return 365


    security.declareProtected(View, 'recueOffsetAsNumberOfDaysForDT')
    def recueOffsetAsNumberOfDaysForDT(self, dt):
		ro = self.getRecueOffset()
		if ro == RO_DAILY: return 1
		if ro == RO_WEEKLY: return 7
		if ro == RO_MONTHLY: return self.numberOfDaysInMonth(dt)
		if ro == RO_YEARLY: return self.numberOfDaysInYear(dt)
		return 0
		
    security.declareProtected(View, 'shiftDateTimeByOffset')
    def shiftDateTimeByOffset(self, dt):
		return dt + self.recueOffsetAsNumberOfDaysForDT(dt)

    #
    # calculators for recue indexes for fast search via portal_catalog
    # ******************************************************************

    security.declareProtected(View, 'calc_recueDays')
    def calc_recueDays(self): 
        """generates list of recue days"""
        rm = self.getRecueMode()
       	if rm == RM_ONE_TIME: return [DT2m(self.startDate)]
       	if rm == RM_UNTIL: return self._genDays(self.startDate, self.getRecueUntilDate())
       	if rm == RM_FOREVER: return self._genDays(self.startDate, FROZEN_DATE_IN_FUTURE)
       	return None
       	
    security.declareProtected(View, 'calc_recueWeeks')
    def calc_recueWeeks(self):
        """
        generates list of recue weeks, 
        recue week is represented by the earliestTime of the first DOW
        """
        rm =  self.getRecueMode()
        if rm == RM_ONE_TIME: return [DT2m(self._genEarliestDOW(self.startDate))]
       	if rm == RM_UNTIL: return self._genWeeks(self.startDate, self.getRecueUntilDate())
       	if rm == RM_FOREVER: return self._genWeeks(self.startDate, FROZEN_DATE_IN_FUTURE)
       	return None

    security.declareProtected(View, 'calc_recueMonths')
    def calc_recueMonths(self):
        """
        generates list of recue months, 
        recue month is represented by the earliestTime of the first DOM
        """
        rm =  self.getRecueMode()
       	if rm == RM_ONE_TIME: return [DT2m(self._genEarliestDOM(self.startDate))]
       	if rm == RM_UNTIL: return self._genMonths(self.startDate, self.getRecueUntilDate())
       	if rm == RM_FOREVER: return self._genMonths(self.startDate, FROZEN_DATE_IN_FUTURE)
       	return None

    security.declareProtected(View, 'calc_recueYears')
    def calc_recueYears(self):
        """
        generates list of recue year, 
        recue year is represented by the earliestTime of the first DOY
        """
        rm =  self.getRecueMode()
       	if rm == RM_ONE_TIME: return [DT2m(self._genEarliestDOY(self.startDate))]
       	if rm == RM_UNTIL: return self._genYears(self.startDate, self.getRecueUntilDate())
       	if rm == RM_FOREVER: return self._genYears(self.startDate, FROZEN_DATE_IN_FUTURE)
       	return None

    #
    # private helper methods for index calculators
    # ******************************************************************
       	       	
    def _genEarliestDOW(self, dt):
	    fdow = dt - dt.dow()
	    return fdow.earliestTime()
	    
    def _genEarliestDOM(self, dt):
	    fdom = dt - (dt.day() - 1)
	    return fdom.earliestTime()
	    
    def _genEarliestDOY(self, dt):
	    fdoy = dt - (dt.dayOfYear() - 1)
	    return fdoy.earliestTime()
	    
	
    #
    # private index values generators
    # ******************************************************************

    def _genDays(self, start_dt, end_dt):
       days = []
       idt = start_dt
       while(idt.lessThanEqualTo(end_dt)):
          tdt = idt.earliestTime()
          days.append(DT2m(tdt))
          idt = self.shiftDateTimeByOffset(tdt).earliestTime()
       return days  

    def _genWeeks(self, start_dt, end_dt):
       # could be optimized by calculate number of shifts 
       # to be realized in a week and Than multiplying before shifting
       weeks = []
       idt = start_dt
       offset = self.getRecueOffset() # little optimalization
       self.setRecueOffset(RO_WEEKLY)
       while(idt.lessThanEqualTo(end_dt)):
          weeks.append(DT2m(self._genEarliestDOW(idt)))
          idt = self.shiftDateTimeByOffset(idt).earliestTime()
       self.setRecueOffset(offset)
       return weeks  
 
    def _genMonths(self, start_dt, end_dt):
       # could be optimized by calculate number of shifts 
       # to be realized in a month and then multiplying before shifting
       months = []
       idt = start_dt
       offset = self.getRecueOffset()	# little optimalization
       self.setRecueOffset(RO_MONTHLY)	
       while(idt.lessThanEqualTo(end_dt)):
          months.append(DT2m(self._genEarliestDOM(idt)))
          idt = self.shiftDateTimeByOffset(idt).earliestTime()
       self.setRecueOffset(offset)
       return months       

    def _genYears(self, start_dt, end_dt):
       # could be optimized by calculate number of shifts 
       # to be realized in year and Than multiplying before shifting
       years = []
       idt = start_dt
       offset = self.getRecueOffset()	# little optimalization
       self.setRecueOffset(RO_YEARLY)
       while(idt.lessThanEqualTo(end_dt)):
          years.append(DT2m(self._genEarliestDOY(idt)))
          idt = self.shiftDateTimeByOffset(idt).earliestTime()
       self.setRecueOffset(offset)
       return years


    #
    # helper methods for attendees' list management
    # ******************************************************************
       		
    security.declareProtected(View, 'hasAttendees')
    def hasAttendees(self):
        """
        returns true if event has any attendee, false elswere 
        """
        return len(self.getAttendees()) > 0
		
		
    security.declareProtected(View, 'isSharedOnlyWithAttendees')
    def isSharedOnlyWithAttendees(self):
		"""
		returns true if that event is only shared with attendees
		"""
		return self.getOnlyForAttendees()
		
	
    security.declareProtected(ManagePortal, 'listAttendeesAsMembers')
    def listAttendeesAsMembers(self):
       """
       returns list of members created from attendees email list
       """
       
       attendees = []
		
       for attemail in self.getAttendees():

          # use the fastest possible search method optimized by GRUF
          mship = getToolByName(self, 'portal_membership')
          members4attemail = mship.searchForMembers(email = attemail)
       
          # if found use it - there should be only one member for given id ;)
          if members4attemail is not None and len(members4attemail) > 0:
             attendees.extend(members4attemail)
		   
       return attendees     

    #
    # management of private/public scope of event for attendees' list
    # ******************************************************************
           
    security.declareProtected(ManagePortal, 'makeSharedWithAttendees')
    def makeSharedWithAttendees(self):
       """
       manages localRoles and permission to share that event only with attendees
       """ 
    
       # permit view only to attendees and owners
       self.manage_permission(View, [ATTENDEE_ROLE, OWNER_ROLE], 0)
		
       # add local roles to attendees who are members!
       for attendee in self.listAttendeesAsMembers():
           self.manage_addLocalRoles(attendee.getId(), [ATTENDEE_ROLE,])
 

    #
    # methods to sending notifications to attendees via e-mail
    # ******************************************************************
           
    security.declareProtected(ManagePortal, 'sendNotificationToAttendees')
    def sendNotificationToAttendees(self, mail_host, attachment):
       """
       sends notification message to all specified attendees
	   """

       for attemail in self.getAttendees():
          msg = self.prepareNotificationForAttendee(attemail, attachment)
          if msg is not None:
             mail_host.send(msg)
                
                
    security.declareProtected(ManagePortal, 'prepareNotificationForAttendee')                
    def prepareNotificationForAttendee(self, attemail, attachment):
       """
       prepares body of a notification message with iCal/vCal files as attachments
       """
              
       owner_name, owner_email = self._getOwnerInfo()
       map = {
          'from': '"%s" <%s>' % (owner_name, owner_email),
          'to': attemail,
          'title': self.Title(),
          'start': self.startDate,
          'body':  attachment
       }
       
       # TODO: iCal/vCal attachement
      
       return NOTIFY_MSG % map
          
          
    #
    # methods to generate iCal/vCal files for e-mail attachments
    # ******************************************************************
    def n2rn(self, s):
       return s.replace('\n', '\r\n')   
   
    def vformat(self, s):
       return s.replace('\r', '\\n').replace('\n', '\\n')
    
    #
    # _attachment is att.body generator parametrized by heade,footr, 
    # and method for generating events body
    #
    def _attachment(self, header, footer, body_generator):
        out = StringIO()
        out.write(header % { 'prodid' : PRODID, })
        out.write(self[body_generator]())
        out.write(footer)
        return self.n2rn(out.getvalue())   
       
    security.declareProtected(View, 'vcs_attachment')
    def vcs_attachment(self):
        """vCalendar output as attachment
        """
        return self._attachment(VCS_HEADER, VCS_FOOTER, 'getVCal')
        
    security.declareProtected(View, 'ics_attachment')
    def ics_attachment(self):
        """iCalendar output as attachment
        """
        return self._attachment(ICS_HEADER, ICS_FOOTER, 'getICal')
        
    
    def _getOwnerInfo(self):
        owner = self.getOwner()
        utool = getToolByName(self, 'portal_url')
        portal = utool.getPortalObject()
        
        try:
            if owner not in [None, SpecialUsers.nobody]:
               mship = getToolByName(self, 'portal_membership')
               for member in mship.searchForMembers(name = owner.getId()):
                  return (
                     member.getProperty('fullname', owner.getId()), 
                     member.getProperty('email', portal.getProperty('email_from_address'))
                  )
        except AttributeError:
            print 'warning: could not determin owner email address, will use ADMIN_EMAIL instead'
            
        return (portal.getProperty('email_from_name'), portal.getProperty('email_from_address'))

    #
    # _getCal is event body generator parametrized by start and end section
    #
    def _getCal(self, event_start, event_end):
        out = StringIO()
        
        # minimal and standard information
        map = {
            'dtstamp'   : DateTime().toZone('UTC').strftime(DATE),
            'created'   : DateTime(self.CreationDate()).toZone('UTC').strftime(DATE),
            'uid'       : self.UID(),
            'modified'  : DateTime(self.ModificationDate()).toZone('UTC').strftime(DATE),
            'summary'   : self.Title(),
            'startdate' : self.start().toZone('UTC').strftime(DATE),
            'enddate'   : self.end().toZone('UTC').strftime(DATE),
            }
        out.write(event_start % map)
        
        # conditional information for recurring event
        if self.getRecueMode() in [RM_UNTIL, RM_FOREVER]:
            # we need to add additional information about recurring
            freq = self.getRecueOffset().upper()
            until = ''
            
            # setup proper until date, in 'forever' case we specify nothing
            if self.getRecueMode() == RM_UNTIL:
               until = ';UNTIL=%s' % self.getRecueUntilDate().toZone('UTC').strftime(DATE)
               
            out.write('RRULE:FREQ=%s;INTERVAL=1%s\n' % (freq, until))
        
        description = self.Description()
        if description:
            out.write('DESCRIPTION:%s\n' % self.vformat(description))
            
        location = self.getLocation()
        if location:
            out.write('LOCATION:%s\n' % self.vformat(location))
            
        eventType = self.getEventType()
        if eventType:
            try:
                out.write('CATEGORIES:%s\n' % str(eventType))
            except:
                import pdb;pdb.set_trace()
                raise
            
        #inserting information about member who organizes event
        owner_name, owner_email = self._getOwnerInfo()
        out.write('ORGANIZER;CN="%s":MAILTO=%s\n' % (owner_name, owner_email))
                
        #inserting information about attendees
        if self.hasAttendees():
           for attemail in self.getAttendees():
              out.write('ATTENDEE;ROLE=REQ-PARTICIPANT:MAILTO:%s\n' % attemail)
        
        out.write(event_end)
        return out.getvalue()    
       
    security.declareProtected(View, 'getICal')
    def getICal(self):
        """get iCal data with recurring and attendees information
        """
        return self._getCal(ICS_EVENT_START, ICS_EVENT_END)

    security.declareProtected(View, 'getVCal')
    def getVCal(self):
        """get vCal data with recurring and attendees information
        """
        return self._getCal(VCS_EVENT_START, VCS_EVENT_END)

    def _viewCal(self, REQUEST, RESPONSE, content_type, ext, header, footer, body_generator):
        RESPONSE.setHeader('Content-Type', content_type)
        RESPONSE.setHeader('Content-Disposition', 'attachment; filename="%s.%s"' % (self.getId(), ext))
        return self._attachment(header, footer, body_generator)
    
    security.declareProtected(View, 'vcs_view')
    def vcs_view(self, REQUEST, RESPONSE):
        """vCalendar output
        """
        return self._viewCal(REQUEST, RESPONSE, 'text/x-vCalendar', 'vcs', VCS_HEADER, VCS_FOOTER, 'getVCal')
                
    security.declareProtected(View, 'ics_view')
    def ics_view(self, REQUEST, RESPONSE):
        """iCalendar output
        """
        return self._viewCal(REQUEST, RESPONSE, 'text/calendar', 'ics', ICS_HEADER, ICS_FOOTER, 'getICal')

                
    #
    # still TODO: 
    # ******************************************************************
    # 1. learn about iCal/vCal format for recurring aspect
    # 2. fixing LAST_DAY_OF_MONTH (LDOM) problem
    # 3. owner e-mail aquisition to be used in FROM message field
                   
		   
registerType(RecurringEvent)
